﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;

using Emgu.CV;
#if !IOS
using Emgu.CV.GPU;
#endif
using Emgu.CV.Structure;

namespace TestProject
{
    /// <summary>
    /// 얼굴 검출 헬퍼
    /// </summary>
    public static class FaceDetectionHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 검출하기 - Detect(image, faceClassifierFilePath, eyeClassifierFilePath, faceRectangleList, eyeRectangleList, detectionTime)

        /// <summary>
        /// 검출하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="faceClassifierFilePath">얼굴 분류기 파일 경로</param>
        /// <param name="eyeClassifierFilePath">눈 분류기 파일 경로</param>
        /// <param name="faceRectangleList">얼굴 사각형 리스트</param>
        /// <param name="eyeRectangleList">눈 사각형 리스트</param>
        /// <param name="detectionTime">검출 시간</param>
        public static void Detect(Image<Bgr, Byte> image, string faceClassifierFilePath, string eyeClassifierFilePath, List<Rectangle> faceRectangleList, List<Rectangle> eyeRectangleList, out long detectionTime)
        {
            Stopwatch stopWatch;

#if !IOS
            if(GpuInvoke.HasCuda)
            {
                using(GpuCascadeClassifier faceGpuCascadeClassifier = new GpuCascadeClassifier(faceClassifierFilePath))
                {
                    using(GpuCascadeClassifier eyeGpuCascadeClassifier = new GpuCascadeClassifier(eyeClassifierFilePath))
                    {
                        stopWatch = Stopwatch.StartNew();

                        using(GpuImage<Bgr, Byte> gpuImage = new GpuImage<Bgr, byte>(image))
                        {
                            using(GpuImage<Gray, Byte> gpuGray = gpuImage.Convert<Gray, Byte>())
                            {
                                Rectangle[] faceRectangleArray = faceGpuCascadeClassifier.DetectMultiScale(gpuGray, 1.1, 10, Size.Empty);

                                faceRectangleList.AddRange(faceRectangleArray);

                                foreach(Rectangle faceRectangle in faceRectangleArray)
                                {
                                    using(GpuImage<Gray, Byte> faceGpuImage = gpuGray.GetSubRect(faceRectangle))
                                    {
                                        using(GpuImage<Gray, Byte> cloneGpuImage = faceGpuImage.Clone())
                                        {
                                            Rectangle[] eyeRectangleArray = eyeGpuCascadeClassifier.DetectMultiScale(cloneGpuImage, 1.1, 10, Size.Empty);

                                            foreach(Rectangle eyeRectangle in eyeRectangleArray)
                                            {
                                                Rectangle temporaryEyeRectangle = eyeRectangle;

                                                temporaryEyeRectangle.Offset(faceRectangle.X, faceRectangle.Y);

                                                eyeRectangleList.Add(temporaryEyeRectangle);
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        stopWatch.Stop();
                    }
                }
            }
            else
#endif
            {
                using(CascadeClassifier faceCascadeClassifier = new CascadeClassifier(faceClassifierFilePath))
                {
                    using(CascadeClassifier eyeCascadeClassifier = new CascadeClassifier(eyeClassifierFilePath))
                    {
                        stopWatch = Stopwatch.StartNew();

                        using(Image<Gray, Byte> grayImage = image.Convert<Gray, Byte>())
                        {
                            grayImage._EqualizeHist();

                            Rectangle[] faceRectangleArray = faceCascadeClassifier.DetectMultiScale
                            (
                                grayImage,
                                1.1,
                                10,
                                new Size(20, 20),
                                Size.Empty
                            );

                            faceRectangleList.AddRange(faceRectangleArray);

                            foreach(Rectangle faceRectangle in faceRectangleArray)
                            {
                                grayImage.ROI = faceRectangle;

                                Rectangle[] eyeRectangleArray = eyeCascadeClassifier.DetectMultiScale
                                (
                                    grayImage,
                                    1.1,
                                    10,
                                    new Size(20, 20),
                                    Size.Empty
                                );

                                grayImage.ROI = Rectangle.Empty;

                                foreach(Rectangle eyeRectangle in eyeRectangleArray)
                                {
                                    Rectangle temporaryEyeRectangle = eyeRectangle;

                                    temporaryEyeRectangle.Offset(faceRectangle.X, faceRectangle.Y);

                                    eyeRectangleList.Add(temporaryEyeRectangle);
                                }
                            }
                        }

                        stopWatch.Stop();
                    }
                }
            }

            detectionTime = stopWatch.ElapsedMilliseconds;
        }

        #endregion
    }
}